#ifndef BUSSIM_PAX_H_INCLUDED
#define BUSSIM_PAX_H_INCLUDED

#include <string>

// added 13.05.2022

// this file contains the definition of the bussim_pax-object to represent individual trips through the simulated
// network.
//
// We define the following object class templates here
// - BUSSIM_PAX => to represent an individual traveller
// - BUSSIM_PAX_LIST => to represent lists of travellers (for several purposes)
// - BUSSIM_PAX_LISTITEM => to place a passenger in a passenger list
//
// - BUSSIM_TP => to represent the travel path of a passenger through the network (seq. of arcs)
// - BUSSIM_TP_SEG => represents a single step in a travel path

class BUSSIM_PAX {

    friend class BUSSIM_PAX_LIST;
    friend class BUSSIM_PAX_LIST_ITEM;

    // this is the template to realize an agent that represent a traveling passenger
public:
    BUSSIM_PAX();           // the standard constructor
    BUSSIM_PAX(long _ID, std::string _PAXNAME, double _SHOW_UP_TIME, long _ORIGIN_NODE, long _DESTINATION_NODE);
    ~BUSSIM_PAX();          // the standard destructor

    void print(void);       // print to shell

    // to do configure at any time
    void specify(long _ID, std::string _PAXNAME, double _SHOW_UP_TIME, long _ORIGIN_NODE, long _DESTINATION_NODE);

    // returns the time when this pax shows up in the system to start its journey
    double get_show_up_time(void);

    // get shortest OD-PATH
    void get_shortest_od_path_dijkstra(class BUSSIM_NETWORK *_NET);

    // display the path in the map
    void display_on_screen(class BUSSIM_NETWORK *_NET);

    // shift the reference to the currently used travel path segment to the next segment
    void proceed_on_travelpath(class BUSSIM_NETWORK *_NET);

    // updates position of walking pax on transfer arc
    void update_pax_pos_walking(class BUSSIM_NETWORK *_NET);


//private:
    long ID;                // a unique identifier of a pax

    std::string PAXNAME;    // to give the passenger a name ...

    int Current_List;       // represents the association to one of the three pax_list
                            // BUSSIM_PAX_UNKNOWN=0; BUSSIM_PAX_IN=1; BUSSIM_PAX_ARRIVED=2;

    int Current_Activity;   // describes the type of the current activity of this passenger
                            // BUSSIM_PAX_WAITING=1;
                            // BUSSIM_PAX_RIDING=2;
                            // BUSSIM_PAX_WALKING=3;
                            // BUSSIM_PAX_NOTHING=0;

    double Show_Up_Time;    // the time when state transitions from BUSSIM_PAX_UNKNOWN to BUSSIM_PAX_IN

    double VELOCITY;        // speed of passenger when walking

    long Origin_Node;       // ID of trip starting node
    long Dest_Node;         // ID of trip terminating node

    class BUSSIM_TP *TravelPath;    // refers to the travelpath of this pax

    class BUSSIM_TP_SEG *Ref2CurrentTPSeg;  // directs to the current travel path segment

    class BUSSIM_PAX_LISTITEM *Ref2WalkingPaxList;      // used to identify this PAX in the list of WALKING paxes
    class BUSSIM_PAX_LISTITEM *Ref2PaxInList;       // used to identify this pax in the list PIX_IN
    class BUSSIM_PAX_LISTITEM *Ref2PointWaitingList;    // used to identify this PAX in a walking waiting list at a point

    long WalkingStartPlatform;      // this is where the current walk starts from

    class BUSSIM_VEHICLE *Ref2Vehicle;          // refers to the vehicle with which the pax is currently riding
};

class BUSSIM_PAX_LISTITEM {
    // this is the template to create passenger lists by connected list items

    friend class BUSSIM_PAX;
    friend class BUSSIM_PAX_LIST;

public:
    class BUSSIM_PAX *get_ref2pax(void);
    class BUSSIM_PAX_LISTITEM *get_previous(void);
    class BUSSIM_PAX_LISTITEM *get_next(void);

private:
    class BUSSIM_PAX *ref2pax;
    class BUSSIM_PAX_LISTITEM *previous;
    class BUSSIM_PAX_LISTITEM *next;
};

class BUSSIM_PAX_LIST {

    friend class BUSSIM_PAX;
    friend class BUSSIM_PAX_LIST_ITEM;

public:
    BUSSIM_PAX_LIST();      // the standard constructor
    BUSSIM_PAX_LIST(std::string _LISTNAME);  // an overloaded constructor
    ~BUSSIM_PAX_LIST();     // the standard destructor

    void append_pax(class BUSSIM_PAX *_REF2PAX, int _MARKER);     // append a passenger
    void print(void);                                // print list to console
    long get_pax_num(void);                          // returns number of paxes in list
    class BUSSIM_PAX *get_last_pax(void);            // returns reference to last pax in the list
    class BUSSIM_PAX *get_first_pax(void);           // returns reference to first pax in the list
    class BUSSIM_PAX_LISTITEM *get_last_pax_item(void);       // returns reference to last item in the list
    class BUSSIM_PAX_LISTITEM *get_first_pax_item(void);      // returns reference to first item in the list
    class BUSSIM_PAX_LISTITEM *get_next_pax_item(class BUSSIM_PAX_LISTITEM *_Ptr);      // returns reference to next item in the list

    double get_first_show_up_time(void);                // return the showup time of the first passenger in this list
    void extract_item(class BUSSIM_PAX_LISTITEM *_PTR);  // resolves _PTR from paxlist
    void insert_item_after(class BUSSIM_PAX_LISTITEM *_PTR,class BUSSIM_PAX_LISTITEM *_REFPTR);  // resolves _PTR from paxlist
    std::string get_listname(void);                         // returns the verbal list descirption
    void set_listname(std::string _LISTNAME);               // sets the listname
    class BUSSIM_SEGMENT *get_ref2schedule_seg(void);        // returns reference to a schedule segment
    void set_ref2schedule_seg(class BUSSIM_SEGMENT *_REF2SCHEDULE_SEG);        // setsreference to a schedule segment

private:
    std::string LISTNAME;        // the description of the list
    long PAXES;             // the number of paxes
    class BUSSIM_PAX_LISTITEM *first_pax_item;  // ref to first stored pax item
    class BUSSIM_PAX_LISTITEM *last_pax_item;   // ref to last stored pax item
    class BUSSIM_SEGMENT *ref2schedule_seg; // refers to a schedule segment
};

class BUSSIM_TP{
    // use this class template to host a BUSSIM_TP_SEG_Sequence representing a travel path of a BUSSIM_PAX

    friend class BUSSIM_PAX;

public:
    BUSSIM_TP();
    BUSSIM_TP(std::string _DESC, class BUSSIM_PAX *_REF2PAX);
    ~BUSSIM_TP();
    void append_segment(class BUSSIM_TP_SEG *_SEG);
    void insert_as_first_segment(class BUSSIM_TP_SEG *_SEG);
    void insert_before(class BUSSIM_TP_SEG *_RefPtr, class BUSSIM_TP_SEG *_OpPtr);
    void dissolve_segment(class BUSSIM_TP_SEG *_OpPtr);
    void print_on_console(class BUSSIM_NETWORK *_NET);
    void print_on_screen(class BUSSIM_NETWORK *_NET);
    long get_first_line(void);

    long get_deboard_node_of_ride(class BUSSIM_TP_SEG *_RefPtr, class BUSSIM_NETWORK *_NET);

private:
    long SEGMENTS;
    std::string DESC;
    long FIRST_LINE;

    class BUSSIM_PAX *REF2PAX;

    class BUSSIM_TP_SEG *REF2FIRST_TP_SEG;
    class BUSSIM_TP_SEG *REF2LAST_TP_SEG;
};

class BUSSIM_TP_SEG {

    friend class BUSSIM_TP;
    friend class BUSSIM_PAX;
    //friend class BUSSIM_NETWORK;

    // we use this class template to describe one step in the OD-path of a pax
public:
    BUSSIM_TP_SEG();        // this is the standard constructor
    BUSSIM_TP_SEG(long _Activity, long _Line, double _Enter_Time, double _Leave_Time, class BUSSIM_ARC *_REF2ARC, class BUSSIM_POINT *_REF2POINT);        // this is the standard constructor
    ~BUSSIM_TP_SEG();       // this is the standard destructor
    long get_line(void);    // returns the line associated with this segment
    class BUSSIM_ARC *get_ref2arc(void);
    class BUSSIM_POINT *get_ref2point(void);
    double get_pos(void);
    long get_activity(void);    // return the activity indicator
    class BUSSIM_VEHICLE *get_ref2vehicle(void);
    class BUSSIM_TP_SEG *get_next(void);
    class BUSSIM_TP_SEG *get_previous(void);
    void set_ref2vehicle(class BUSSIM_VEHICLE *_PtrVeh);

private:
    long Activity;
    long Enter_Point;
    long Leave_Point;
    long Line;      //(-1 = walking, >=0 lines)

    double POS; // [0-1] to represent position of walking pax on a transfer arc

    double Enter_Time;
    double Leave_Time;

    class BUSSIM_ARC *REF2ARC;
    class BUSSIM_POINT *REF2POINT;

    class BUSSIM_VEHICLE *REF2VEHICLE;

    class BUSSIM_TP_SEG *previous;
    class BUSSIM_TP_SEG *next;
};

#endif // BUSSIM_PAX_H_INCLUDED
